﻿using System;
using AITB.Core;
using AITB.Helpers;
using AITB.Settings;
using OPT.Settings;
using System.Windows.Forms;
using AITB.Solvers;
using System.Collections.Generic;
using System.Linq;

namespace OPT.Solvers.GWO
{
    public class GreyWolfOptimizer : GenericSolver
    {
        private Problem _problem;
        private Chromosome[] _population;

        private double[] _alfa;
        private double[] _beta;
        private double[] _delta;
        private double _alfaScore=Double.MaxValue, _betaScore=double.MaxValue, _deltaScore=double.MaxValue;

        public GreyWolfOptimizer():base()
        {
        }

        public override void Solve(Problem problem)
        {
            _problem = problem;

            _alfa = new double[_problem.TotalDimensionality];
            _beta = new double[_problem.TotalDimensionality];
            _delta = new double[_problem.TotalDimensionality];

            _alfa.InitializeWithZero();
            _beta.InitializeWithZero();
            _delta.InitializeWithZero();

            _population = Chromosome.InitializePopulation(_problem);

            double a;
            for (int i = 0; i < OptimizerSettings.Generations; i++)
            {
                a = 2.0 - i * (2.0 / OptimizerSettings.Generations);

                UpdatePopulation(a);

                if (FesValues.Contains(GeneralSettings.ActualFes))
                {
                    Solution = (Chromosome)_population[_population.GetIndexOfBestSolution()].Clone();
                    Solution.WriteSolutionToFile(GeneralSettings.ActualFes);
                }

                //upgrade the best solution and save it
                var index = _population.GetIndexOfBestSolution();
                _population[index].ImproveEntity(problem);
                Solution = (Chromosome)_population[index].Clone();

                Solution.WriteSolutionToFile();

                if (OptimizerSettings.PerformanceMetrics)
                    _population.DeterminePerformanceMetrics(i, ref CurrentPopulationDiversity);

                Displayer.AppendText("" + (OptimizerSettings.Generations - i) + " ");
                Application.DoEvents();

                if (i > OptimizerSettings.Generations || GeneralSettings.StopTraining ||
                    Solution.Fitness == 1e+10 || GeneralSettings.ActualFes > OptimizerSettings.MaximFes)
                    i = OptimizerSettings.Generations;
            }           
        }

        private void UpdatePopulation(double a)
        {
            for (int ind = 0; ind < OptimizerSettings.PopSize; ind++)
            { 
                for (int j = 0; j < _problem.TotalDimensionality; j++)
                {
                    _population[ind].Data[j] = (ComputeXForPosition(a, _alfa, ind, j) + ComputeXForPosition(a, _beta, ind, j) + ComputeXForPosition(a, _delta, ind, j)) / 3.0;

                }
                _problem.VerifyBoundaries(_population[ind]);
                _population[ind].ComputeFitness(ref _problem);

                if (_population[ind].Fitness > _alfaScore)
                {
                    _alfaScore = _population[ind].Fitness;
                    _alfa =(double[]) _population[ind].Data.Clone();
                }

                if (_population[ind].Fitness < _alfaScore && _population[ind].Fitness > _betaScore)
                { 
                    _betaScore= _population[ind].Fitness;
                    _beta = (double[])_population[ind].Data.Clone();
                }

                if (_population[ind].Fitness < _alfaScore && _population[ind].Fitness < _betaScore && _population[ind].Fitness > _deltaScore)
                {
                    _deltaScore = _population[ind].Fitness;
                    _delta = (double[])_population[ind].Data.Clone();
                }

            }
        }

        private double ComputeXForPosition(double a, double[] position_vect, int individual, int position)
        {
            double r1 = ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng);
            double r2 = ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng);

            double A1 = 2.0 * a * r1 - a; //Eq 3.3
            double C1 = 2.0 * r2; //Eq. 3.4

            double D_alpha = Math.Abs(C1 * position_vect[position] - _population[individual].Data[position]); //eq. 3.5
            double X = position_vect[position] - A1 * D_alpha; //eq 3.6

            return X;
        }

       

        public override string GetControlParameters()
        {
            throw new NotImplementedException();
        }
    }
}

