﻿using AITB.Core;
using AITB.Helpers;
using AITB.Settings;
using OPT.Settings;
using System;
using System.Linq;
using System.Windows.Forms;

namespace AITB.Solvers.BFO
{
    public class BacterialForagingOptimization : GenericSolver
    {
        //ref 6660

        private Bacterium[] colony;


        public override void Solve(Problem problem)
        {
            var _delSols = 0;

            var Nc = 20;
            // chemotactic steps, index = j

            var Ns = 5;
            // maximum swim steps, index = m

            var Nre = 8;
            // reproduction steps, index = k

            var Ned = OptimizerSettings.Generations;
            //20; //4;
            // dispersal steps, index = l ('el')

            var Ped = 0.25;
            // probability of dispersal

            var Ci = 0.05;
            // basic bacteria movement increment size (for all bacteria)


            // initialize bacteria colony
            colony = new Bacterium[OptimizerSettings.PopSize];
            // initialize a colony of bacteria at random positions (but not the costs)

            double minFitness = double.MaxValue;
            double maxFitness = double.MinValue;

            // costs and identification of best 
            var bestCost = double.MinValue;
            var indexOfBest = 0;

            for (var i = 0; i < OptimizerSettings.PopSize; i++)
            {
                colony[i] = new Bacterium(problem.Parameters);


                // compute cost
                colony[i].ComputeFitness(ref problem);
                colony[i].prevCost = colony[i].Fitness;

                //best
                if (colony[i].Fitness > bestCost)
                {
                    bestCost = colony[i].Fitness;
                    indexOfBest = i;
                }
            }


            Solution = (Bacterium)colony[indexOfBest].Clone();


            // main processing
            var t = 0;
            // time counter
            // eliminate-disperse loop
            for (var l = 0; l < Ned; l++)
            {
                // reproduce-eliminate loop
                for (var k = 0; k < Nre; k++)
                {
                    // chemotactic loop; the lifespan of each bacterium
                    for (var j = 0; j < Nc; j++)
                    {
                        // reset the health of each bacterium to 0.0 
                        for (var x = 0; x < OptimizerSettings.PopSize; x++)
                            colony[x].health = 0.0;

                        var i = 0;
                        // each bacterium
                        while (i < OptimizerSettings.PopSize)
                        {
                            var tumble = new double[problem.Parameters.EncodingInfo.NoGenes];
                            // tumble (point in a new direction)
                            // (hi - lo) * r + lo => random i [-1, +1]
                            //includes Rng parameters
                            if (OptimizerSettings.AdaptationOfInitializationParameters)
                            {
                                for (var p = 0; p < problem.TotalDimensionality; p++)
                                    tumble[p] = 2.0 * ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng, Solution.Data[problem.TotalDimensionality - 1]) - 1.0;
                            }
                            else
                            {
                                for (var p = 0; p < problem.TotalDimensionality; p++)
                                    tumble[p] = 2.0 * ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng) - 1.0;
                            }

                            var rootProduct = 0.0;
                            for (var p = 0; p < problem.TotalDimensionality; p++)
                                rootProduct += tumble[p] * tumble[p];
                            rootProduct = Math.Sqrt(rootProduct);

                            if (OptimizerSettings.SelfAdaptationParam)
                                Ci = colony[i].Data[problem.TotalDimensionality - 1];
                            // move in new direction
                            for (var p = 0; p < problem.TotalDimensionality; p++)
                                colony[i].Data[p] += Ci * tumble[p] / rootProduct;

                            colony[i].VerifyExcedingLimits(problem.Parameters.EncodingInfo);
                            // update costs of new position
                            colony[i].prevCost = colony[i].Fitness;
                            colony[i].ComputeFitness(ref problem);



                            if (FesValues.Contains(GeneralSettings.ActualFes))
                            {
                                Solution.WriteSolutionToFile(GeneralSettings.ActualFes);
                            }

                            var m = 0;
                            // swim or not based on prev and curr costs
                            // we are improving
                            while (m < Ns && colony[i].Fitness > colony[i].prevCost)
                            {
                                m += 1;
                                // swim counter
                                // move in current direction
                                for (var p = 0; p < problem.Parameters.EncodingInfo.NoGenes; p++)
                                    colony[i].Data[p] += Ci * tumble[p] / rootProduct;
                                colony[i].prevCost = colony[i].Fitness;
                                // update costs

                                colony[i].VerifyExcedingLimits(problem.Parameters.EncodingInfo);
                                colony[i].ComputeFitness(ref problem);

                                if (FesValues.Contains(GeneralSettings.ActualFes))
                                {
                                    Solution.WriteSolutionToFile(GeneralSettings.ActualFes);
                                }
                            }

                            colony[i].health += colony[i].Fitness;
                            // health is an accumulation of costs during bacterium's life


                            // while improving
                            // did we find a new best?

                            if (colony[i].Fitness > bestCost)
                            {
                                bestCost = colony[i].Fitness;
                                Solution = (Bacterium)colony[i].Clone();
                            }
                            i += 1;
                        }
                        // i, each bacterium in the chemotactic loop

                        t += 1;
                        // increment the time counter

                        if (Solution.Fitness == 1e+10 || GeneralSettings.ActualFes > OptimizerSettings.MaximFes)
                            break;

                    }
                    // j, chemotactic loop

                    // reproduce the healthiest half of bacteria, eliminate the other half
                    var sortedColony = colony.OrderBy(o => o.health).ToArray();
                    Array.Reverse(sortedColony);

                    // sort from smallest health- fighest fitness (best) to highest health-lowest fitness (worst)
                    // left points to a bacterium that will reproduce
                    for (var left = 0; left < OptimizerSettings.PopSize / 2; left++)
                    {
                        var right = left + OptimizerSettings.PopSize / 2;
                        colony[left] = new Bacterium(ref sortedColony[left]);
                        // right points to a bad bacterium in the rigt side of array that will die
                        colony[right] = new Bacterium(ref colony[left]);
                    }

                    if (Solution.Fitness == 1e+10 || GeneralSettings.ActualFes > OptimizerSettings.MaximFes)
                        break;
                }
                // k, reproduction loop

                var avgFitness = 0.0;
                for (int i = 0; i < OptimizerSettings.PopSize; i++)
                {
                    avgFitness += colony[i].Fitness;
                    if (minFitness > colony[i].Fitness) minFitness = colony[i].Fitness;
                    if (maxFitness < colony[i].Fitness) maxFitness = colony[i].Fitness;
                }
                avgFitness = avgFitness / OptimizerSettings.PopSize;

                // eliminate-disperse
                for (var i = 0; i < OptimizerSettings.PopSize; i++)
                {
                    var prob = OptimizerSettings.AdaptationOfInitializationParameters ? ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng, Solution.Data[problem.TotalDimensionality - 1]) : ExtendedRandom.GenerateRandomNumber(OptimizerSettings.Rng);
                    // disperse this bacterium to a random position


                    var Pedd = Ped;
                    ////ref 7133
                    ////var 1
                    //Pedd = Ped * (i - OptimizerSettings.PopSize / 4 + 0.5) * (l - 1) /
                    //       ((Ned - 1) * (OptimizerSettings.PopSize / 2 - 1));

                    ////ref 7133
                    ////var 2
                    //Pedd = Ped + Ped * (l - 1) / 
                    //       (Ned - 1) * (2 * (i - 1) / (OptimizerSettings.PopSize / 2 - 1) * (i - 1) / (OptimizerSettings.PopSize / 2 - 1) - 1);

                    //my adaptive variant
                    if (OptimizerSettings.SelfAdaptationParam) Pedd = avgFitness / 1e+10;

                    if (prob < Pedd)
                    {
                        if (OptimizerSettings.AdaptationOfInitializationParameters)
                        { colony[i] = new Bacterium(problem.Parameters, Solution.Data[problem.TotalDimensionality - 1]); }
                        else { colony[i] = new Bacterium(problem.Parameters); }

                        colony[i].ComputeFitness(ref problem);

                        // new best by pure luck?
                        if (colony[i].Fitness > bestCost)
                        {
                            bestCost = colony[i].Fitness;
                            Solution = (Bacterium)colony[i].Clone();
                            colony[i] = (Bacterium)Solution.ImproveEntity(problem);
                            Solution= (Bacterium)colony[i].Clone();
                        }

                        if (FesValues.Contains(GeneralSettings.ActualFes))
                        {
                            Solution.WriteSolutionToFile(GeneralSettings.ActualFes);
                        }
                    }
                }


                Solution.WriteSolutionToFile();
                if (OptimizerSettings.PerformanceMetrics)
                {
                    colony.DeterminePerformanceMetrics(l, ref CurrentPopulationDiversity);
                }

                Displayer.AppendText("" + (OptimizerSettings.Generations - l) + " ");
                Application.DoEvents();

                if (Solution.Fitness == 1e+10 || GeneralSettings.ActualFes > OptimizerSettings.MaximFes)
                    break;

            }
            // l, elimination-dispersal loop, end processing
        }

        public override string GetControlParameters()
        {
            throw new NotImplementedException();
        }
    }
}